"""
Testing global utilities
"""

import unittest
from functools import partial

import numpy as np

from acia.utils import lut_mapping

# lut function for mapping from [0, 1000] to [0, 255]
lut_func = partial(lut_mapping, in_min=0, in_max=1000, out_min=0, out_max=255)


class TestMappingLUT(unittest.TestCase):
    """Test mapping lut functionality"""

    def test_number_mapping(self):
        """Test raw number mapping"""
        self.assertEqual(lut_func(0), 0)
        self.assertEqual(lut_func(1000), 255)
        self.assertEqual(lut_func(-5), 0)
        self.assertEqual(lut_func(1005), 255)

        self.assertEqual(lut_func(np.array([1]), dtype=np.uint8), 0)

    def test_image_mapping(self):
        """Test full image mapping"""
        image = np.array([[5, 1000], [16000, 5000]], dtype=np.int16)

        mapped_image = lut_func(image, dtype=np.uint8)

        for gt, pred in zip(image.flatten(), mapped_image.flatten()):
            self.assertEqual(
                pred, np.floor(np.clip(gt, 0, 1000) / 1000 * 255).astype(np.uint8)
            )


if __name__ == "__main__":
    unittest.main()



""" Module to convert tracking formats """

from __future__ import annotations

import json
import re
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd

from acia.base import Contour, Overlay


def parse_simple_tracking(file_content: str) -> tuple[Overlay, nx.DiGraph]:
    """Parse simple tracking format from file content string

    Args:
        file_content (str): simple tracking format file content

    Returns:
        Tuple[Overlay, nx.DiGraph]: segmentation overlay and tracking graph
    """

    data = json.loads(file_content)

    segmentation_data = data["segmentation"]
    tracking_data = data["tracking"]

    # deal with the segmentation first
    all_detections = []

    # create contours
    for det in segmentation_data:
        all_detections.append(
            Contour(det["contour"], -1, det["frame"], det["id"], det["label"])
        )

    segmentation_overlay = Overlay(all_detections)

    # deal with the tracking
    tracking = nx.DiGraph()

    tracking.add_nodes_from(map(lambda ov: ov.id, segmentation_overlay))
    for cont in segmentation_overlay:
        tracking.nodes[cont.id]["frame"] = cont.frame

    # create graph from id links
    for link in tracking_data:
        tracking.add_edge(link["sourceId"], link["targetId"])

    return segmentation_overlay, tracking


def gen_simple_tracking(overlay: Overlay, tracking_graph: nx.Graph) -> str:
    """Create a simple tracking format from overlay and tracking graph

    Args:
        overlay (Overlay): segmentation overlay
        tracking_graph (nx.Graph): tracking graph

    Returns:
        str: simple tracking format string
    """

    segmentation_data = []
    for cont in overlay:

        coordinates = cont.coordinates

        if isinstance(coordinates, np.ndarray):
            coordinates = coordinates.tolist()

        segmentation_data.append(
            dict(label=cont.label, contour=coordinates, id=cont.id, frame=cont.frame)
        )

    tracking_data = []
    for edge in tracking_graph.edges:
        tracking_data.append(dict(sourceId=edge[0], targetId=edge[1]))

    simpleTracking = dict(segmentation=segmentation_data, tracking=tracking_data)

    return json.dumps(simpleTracking)


def load_ctc_tracklet_graph(file: Path):
    colnames = ["label", "t_start", "t_end", "parent"]

    ctc_df = pd.read_csv(
        file,
        names=colnames,
        header=None,
        dtype={"label": "int", "t_start": "int", "t_end": "int", "parent": "int"},
        delim_whitespace=" ",
    )

    tracklet_graph = nx.DiGraph()

    for _, row in ctc_df.iterrows():
        label, t_start, t_end, parent_label = row.to_list()
        tracklet_graph.add_node(label, start_frame=t_start, end_frame=t_end)

        if parent_label != 0:
            tracklet_graph.add_edge(parent_label, label)

    return tracklet_graph


def ctc_track_graph(ov: Overlay, tracklet_graph: nx.DiGraph):
    """Computes the ctc track graph (every cell detection is a node) based on cell detections (overlay) and the tracklet graph (every tracklet is one node).

    Hint: overlay labels and tracklet_graph node ids need to align.

    Args:
        ov (Overlay): _description_
        tracklet_graph (nx.DiGraph): _description_

    Returns:
        _type_: _description_
    """

    track_graph = nx.DiGraph()

    # add all the nodes
    for cont in ov:
        track_graph.add_node(cont.id, frame=cont.frame)

    tracklets = {}
    for cont in ov:
        tracklets[cont.label] = tracklets.get(cont.label, []) + [cont]

    for tracklet_label in tracklets:
        tracklets[tracklet_label] = sorted(
            tracklets[tracklet_label], key=lambda c: c.frame
        )

    for tracklet_label, tracklet_nodes in tracklets.items():

        # add tracklet edges
        for contA, contB in zip(tracklet_nodes, tracklet_nodes[1:]):
            track_graph.add_edge(contA.id, contB.id)

        for pred_label in tracklet_graph.predecessors(tracklet_label):
            track_graph.add_edge(tracklets[pred_label][-1].id, tracklet_nodes[0].id)

        for succ_label in tracklet_graph.successors(tracklet_label):
            track_graph.add_edge(tracklet_nodes[-1].id, tracklets[succ_label][0].id)

    return track_graph


def read_ctc_tracking(file: Path) -> list[dict]:
    """Reads a tracking ctc file line by line

    Args:
        file (Path): File to ctc tracking txt file

    Returns:
        list[dict]: List of the individual lines with a dict[id, start_frame, end_frame, parent_id] containing the tracking information
    """

    # compile regex to match the file
    regex = re.compile(
        r"^(?P<id>\d+) (?P<start_frame>\d+) (?P<end_frame>\d+) (?P<parent_id>\d+)"
    )

    with open(str(file), encoding="utf-8") as input_file:
        data = []
        for line in input_file.readlines():
            m = regex.match(line)

            if m is None:
                # no regex match for this line
                continue

            # add the extracted data
            data.append(m.groupdict())

    return data


def tracking_to_graph(data: list[dict]) -> nx.DiGraph:
    """Populates a ctc tracking into a full tracking lineage where every detection has its own node with a unique id based on (ctc_id, frame)

    Args:
        data (list[dict]): Output of :func:`read_ctc_tracking`

    Returns:
        nx.DiGraph: A lineage graph where every detection has its unique node (id, frame) and the edges represent the linking
    """

    graph = nx.DiGraph()

    # go through every ctc line
    for item in data:
        # iterate every frame for that the cell track exists
        for frame in range(int(item["start_frame"]), int(item["end_frame"]) + 1):
            # add node with unique id (ctc_id, frame)
            graph.add_node((item["id"], frame), frame=frame)
            # add non-division links
            if graph.has_node((item["id"], frame - 1)):
                graph.add_edge((item["id"], frame - 1), (item["id"], frame))

    # add division links
    for item in data:
        if int(item["parent_id"]) != 0:

            # get the time it divides
            # split_frame = int(item["start_frame"]) - 1

            # extract source and target
            source_candidates = list(
                # pylint: disable=cell-var-from-loop
                filter(lambda n: n[0] == item["parent_id"], graph.nodes)
            )
            latest_source = np.argmax(list(map(lambda n: n[1], source_candidates)))
            source = source_candidates[
                latest_source
            ]  # (item["parent_id"], split_frame)
            target = (item["id"], int(item["start_frame"]))

            if not graph.has_node(source):
                print("Error")

            assert graph.has_node(source), f"Source: {source}"
            assert graph.has_node(target), f"Target: {target}"

            # add the edge
            graph.add_edge(source, target)

    return graph



""" Functionality for single-cell analysis """

from __future__ import annotations

import logging
import os
import shutil
from collections.abc import Iterable
from functools import reduce
from itertools import starmap
from multiprocessing import Pool
from pathlib import Path

import numpy as np
import pandas as pd
import papermill as pm
from numpy import ma
from pint._typing import UnitLike
from tqdm.auto import tqdm

from acia import Q_, U_
from acia.base import BaseImage, ImageSequenceSource, Overlay
from acia.utils import pairwise_distances

DEFAULT_UNIT_LENGTH = "micrometer"
DEFAULT_UNIT_AREA = "micrometer ** 2"


class PropertyExtractor:
    """Base class for single-cell property extractor"""

    def __init__(
        self, name: str, input_unit: UnitLike, output_unit: UnitLike | None = None
    ):
        self.name = name

        # try to parse input quantity
        self.input_unit = Q_(input_unit)
        if self.input_unit.dimensionless and isinstance(self.input_unit.magnitude, U_):
            # if we have no dimension and magnitude is unit -> we better go with a unit
            self.input_unit = U_(input_unit)
        if output_unit:
            # parse output unit
            self.output_unit = U_(output_unit)
        else:
            # no conversion if no output unit is specified
            self.output_unit = self.input_unit

        # test the conversion here
        self.output_unit.is_compatible_with(self.input_unit)

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        """Extract the desired properties for a single contour

        Args:
            contour (Contour): contour for the qunatity
            overlay (Overlay): overlay containing all contours
            df (pd.DataFrame): DataFrame of properties so far

        Raises:
            NotImplementedError: Please implement this method
        """
        raise NotImplementedError()

    def convert(self, input: float | Q_) -> float:
        """Converts input to the specified output unit

        Args:
            input (float | Quantity): Input value

        Returns:
            float: the magnitude in the output unit
        """
        if isinstance(input, Q_):
            # 1. convert input to input unit
            # 2. scale with input unit
            # 3. convert to output unit
            return (
                (input.to(self.input_unit).magnitude * self.input_unit)
                .to(self.output_unit)
                .magnitude
            )
        else:
            # 1. scale input with input unit/quantity
            # 2. convert to output unit
            return (input * self.input_unit).to(self.output_unit).magnitude


class ExtractorExecutor:
    """Executor to extract a list of single-cell properties from segmentation and images"""

    def __init__(self) -> None:
        self.units = {}

    def execute(
        self, overlay: Overlay, images: list, extractors: list[PropertyExtractor] = None
    ):
        if extractors is None:
            extractors = []

        df = pd.DataFrame()
        for extractor in tqdm(extractors):
            print(f"Extracting: {extractor.name}...")
            result_df, units = extractor.extract(overlay, images, df)

            df = pd.concat([df, result_df], ignore_index=False, sort=False, axis=1)

            self.units.update(**units)

        return df


class AreaEx(PropertyExtractor):
    """Extract area for every contour"""

    def __init__(
        self,
        input_unit: UnitLike | None = DEFAULT_UNIT_AREA,
        output_unit: UnitLike | None = DEFAULT_UNIT_AREA,
    ):
        PropertyExtractor.__init__(
            self, "area", input_unit=input_unit, output_unit=output_unit
        )

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        areas = []
        for cont in overlay:
            areas.append(self.convert(cont.area))

        return pd.DataFrame({self.name: areas}), {self.name: self.output_unit}


class LengthEx(PropertyExtractor):
    """Extracts width of cells based on the shorter edge of a minimum rotated bbox approximation"""

    def __init__(
        self,
        input_unit: UnitLike | None = DEFAULT_UNIT_LENGTH,
        output_unit: UnitLike | None = DEFAULT_UNIT_LENGTH,
    ):
        PropertyExtractor.__init__(
            self, "length", input_unit=input_unit, output_unit=output_unit
        )

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        lengths = []
        for cont in overlay:
            lengths.append(
                self.convert(
                    # longer edge of minimum roated bbox
                    np.max(
                        pairwise_distances(
                            np.array(
                                cont.polygon.minimum_rotated_rectangle.exterior.coords
                            )
                        )
                    )
                )
            )

        return pd.DataFrame({self.name: lengths}), {self.name: self.output_unit}


class WidthEx(PropertyExtractor):
    """Extracts width of cells based on the shorter edge of a minimum rotated bbox approximation"""

    def __init__(
        self,
        input_unit: UnitLike | None = DEFAULT_UNIT_LENGTH,
        output_unit: UnitLike | None = DEFAULT_UNIT_LENGTH,
    ):
        PropertyExtractor.__init__(
            self, "width", input_unit=input_unit, output_unit=output_unit
        )

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        """Extract width information for all contours"""
        widths = []
        for cont in overlay:
            widths.append(
                self.convert(
                    # shorter edge of bbox approximation
                    np.min(
                        # measure edge lengths of bbox approximation
                        pairwise_distances(
                            np.array(
                                # bbox approaximation
                                cont.polygon.minimum_rotated_rectangle.exterior.coords
                            )
                        )
                    )
                )
            )

        return pd.DataFrame({self.name: widths}), {self.name: self.output_unit}


class FrameEx(PropertyExtractor):
    """Extract the frame information for every contour"""

    def __init__(self):
        super().__init__("frame", 1)

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        frames = []
        for cont in overlay:
            frames.append(self.convert(cont.frame))

        return pd.DataFrame({self.name: frames}), {self.name: self.output_unit}


class IdEx(PropertyExtractor):
    """Extract single-cell id for every contour"""

    def __init__(self):
        super().__init__("id", 1)

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        ids = []
        for cont in overlay:
            ids.append(self.convert(cont.id))
        return pd.DataFrame({self.name: ids}), {self.name: self.output_unit}


class LabelEx(PropertyExtractor):
    """Extract single-cell label (from tracking) for every contour"""

    def __init__(self):
        super().__init__("label", 1)

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        labels = []
        for cont in overlay:
            labels.append(self.convert(cont.label))
        return pd.DataFrame({self.name: labels}), {self.name: self.output_unit}


class TimeEx(PropertyExtractor):
    """Extract time information for every contour"""

    def __init__(self, input_unit: UnitLike, output_unit: UnitLike | None = "hour"):
        super().__init__("time", input_unit, output_unit)

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        times = []
        for _, row in df.iterrows():
            times.append(self.convert(row["frame"]))

        return pd.DataFrame({self.name: times}), {self.name: self.output_unit}


class DynamicTimeEx(PropertyExtractor):
    """Extract time information for every contour when timepoints are not equi-distant"""

    def __init__(
        self,
        timepoints: list,
        relative=True,
        input_unit: UnitLike = "second",
        output_unit: UnitLike | None = "hour",
    ):
        super().__init__("time", input_unit, output_unit)

        if len(timepoints) == 0:
            raise ValueError("Need non-empty timepoint list")

        self.timepoints = np.array(timepoints)

        if relative:
            self.timepoints -= self.timepoints[0]

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):

        # get the number of frames
        num_frames = np.unique(df["frame"])

        if len(self.timepoints) != len(num_frames):
            raise ValueError(
                f"Number of specified timepoints does not match with number of frames: {len(num_frames)=} vs. {len(self.timepoints)} timepoints"
            )

        times = []
        for _, row in df.iterrows():
            times.append(
                # convert to timepoint units
                self.convert(
                    # lookup frame timepoint
                    self.timepoints[row["frame"]]
                )
            )

        return pd.DataFrame({self.name: times}), {self.name: self.output_unit}


class PositionEx(PropertyExtractor):
    """Extract cell center information from image RoI detections"""

    def __init__(
        self,
        input_unit: UnitLike,
        output_unit: UnitLike | None = DEFAULT_UNIT_LENGTH,
    ):
        super().__init__("position", input_unit=input_unit, output_unit=output_unit)

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        positions_x = []
        positions_y = []
        for cont in overlay:
            positions_x.append(self.convert(cont.center[0]))
            positions_y.append(self.convert(cont.center[1]))

        return pd.DataFrame({"position_x": positions_x, "position_y": positions_y}), {
            "position_x": self.output_unit,
            "position_y": self.output_unit,
        }


class FluorescenceEx(PropertyExtractor):
    """Extracting fluorescence properties from image sequence and RoI detections"""

    def __init__(
        self,
        channels,
        channel_names,
        summarize_operator=np.median,
        input_unit: UnitLike = "1",
        output_unit: UnitLike | None = "",
        parallel=6,
    ):
        super().__init__("Fluorescence", input_unit=input_unit, output_unit=output_unit)

        self.channels = channels
        self.channel_names = channel_names
        self.summarize_operator = summarize_operator
        self.parallel = parallel

        assert len(self.channels) == len(
            self.channel_names
        ), "Number of channels and number of channel names must comply"

    @staticmethod
    def extract_fluorescence(
        overlay: Overlay,
        image: BaseImage,
        channels: list[int],
        channel_names: list[str],
        summarize_operator,
    ):
        """Extract fluorescence information based on an overlay(segmentation) and corresponding image.

        Args:
            overlay (Overlay): Ovleray providing the image segmentation information
            image (BaseImage): the image itself
            channels (List[int]): list of channels (image channels) we want to investigate
            channel_names (List[str]): list of names for the channel results
            summarize_operator (_type_): summarizing operator, e.g. np.media, to compress all fluorescence values to a single one

        Returns:
            pd.DataFrame: pandas data frame containing columns of channel_names and the rows represent the extracted fluorescence
        """
        channel_values = [[] for _ in channels]

        for cont in overlay:
            for ch_id, channel in enumerate(channels):
                raw_image = image.get_channel(channel)

                height, width = raw_image.shape[:2]

                # draw cell mask
                roi_mask = cont.toMask(height=height, width=width)

                # create masked array
                masked_roi = ma.masked_array(raw_image, mask=~roi_mask)

                # compute fluorescence response
                value = summarize_operator(masked_roi.compressed())

                channel_values[ch_id].append(value)

        return pd.DataFrame(
            {channel_names[i]: channel_values[i] for i in range(len(channels))}
        )

    def extract(self, overlay: Overlay, images: ImageSequenceSource, df: pd.DataFrame):
        assert overlay.numFrames() == len(
            images
        ), "Please make sure that the frames in your overlay fit to the frames in your image source"

        def iterator(timeIterator):
            for i, overlay in enumerate(timeIterator):
                yield (
                    overlay,
                    images.get_frame(i),
                    self.channels,
                    self.channel_names,
                    self.summarize_operator,
                )

        result = None

        if self.parallel > 1:
            try:
                with Pool(self.parallel) as p:
                    result = p.starmap(
                        FluorescenceEx.extract_fluorescence,
                        iterator(overlay.timeIterator()),
                        chunksize=5,
                    )

            except Exception as e:
                logging.error(
                    "Parallel fluorescence extraction failed! Please run with 'parallel=1' to investigate the error!"
                )
                raise e

        else:
            result = starmap(
                FluorescenceEx.extract_fluorescence, iterator(overlay.timeIterator())
            )

        # concatenate all results
        result = reduce(lambda a, b: pd.concat([a, b], ignore_index=True), result)

        return result, {
            self.channel_names[i]: self.output_unit for i in range(len(self.channels))
        }


def scale(
    output_path: Path,
    analysis_script: Path | list[Path],
    image_ids: list[int],
    additional_parameters=None,
    exist_ok=False,
    execution_naming=lambda image_id: f"execution_{image_id}",
    exist_skip=False,
):
    """Scale an analysis notebook to several image sequences

    **Hint:** the analysis script should only use absolute paths as the file is copied and executed in another folder.

    Args:
        output_path (Path): the general output path to the storage
        analysis_script (Path): the template script
        image_ids (List[int]): list of (OMERO) image sources
        additional_parameters (dict): Parameters to be inserted into the jupyter script
        exist_ok (Bool): True when it is okay that the directory exists, False will throw an error when the directory exists.
        exist_skip (Bool): If true existing executions are skipped.
    """

    if isinstance(analysis_script, str):
        # if this is just a single string, then we make it a list of a single path
        analysis_script = [Path(analysis_script)]
    elif isinstance(analysis_script, Path):
        analysis_script = [analysis_script]
    elif isinstance(analysis_script, Iterable):
        analysis_script = list(map(Path, analysis_script))

    for script in analysis_script:
        if not script.exists():
            raise ValueError(f"Analysis script {script} does not exist!")

    if additional_parameters is None:
        additional_parameters = {}

    experiment_executions = []

    failed_ids = []

    failed_ids = []

    for image_id in tqdm(image_ids):

        try:

            # create the main output folder
            output_parent = output_path / execution_naming(image_id)
            os.makedirs(output_parent, exist_ok=exist_ok)

            for script in analysis_script:
                # path to the new notebook file
                # every execution should have its own folder to store local files
                output_file = output_parent / script.name

                if output_file.exists() and exist_skip:
                    # the notebook exists and we should skip it
                    continue

                shutil.copy(script, output_file)

                # parameters to integrate into notebook
                parameters = dict(
                    storage_folder=str(output_file.parent.absolute()),
                    image_id=image_id,
                    **additional_parameters,
                )

                # execute the notebook
                pm.execute_notebook(
                    output_file,
                    output_file,
                    parameters=parameters,
                    cwd=output_file.parent,
                )

                # save experiment in list
                experiment_executions.append(
                    dict(parameters=parameters, storage_folder=output_file.parent)
                )
        except pm.PapermillExecutionError:
            failed_ids.append(image_id)

    if len(failed_ids) > 0:
        error_ratio = len(failed_ids) / len(image_ids) * 100

        logging.warning(
            "The scaling failed in %d/%d (%.3f%%) executions. Please report failes with the link to the script and the image id to your administrator in order to further improve the software.",
            len(failed_ids),
            len(image_ids),
            error_ratio,
        )
        if error_ratio > 10:
            # error rates of more than 10% are definitively acceptable
            logging.error("Such a high error rate is not acceptable!")

    return experiment_executions



""" Local segmentation functionality dealing with files from HDD.
"""

import os
import os.path as osp

import cv2
import numpy as np
import roifile
import tifffile

from acia.base import BaseImage, Contour, ImageSequenceSource, Overlay, RoISource


def prepare_image(image, normalize_image=True):
    """Normalize and convert image to RGB.

    Args:
        image ([type]): [description]
        normalize_image (bool, optional): Whether to normalize the image into uint8 domain (0-255). Defaults to True.
    Returns:
        [np.array]: RGB image (Width, height, 3 color channels)
    """
    # normalize image space
    if normalize_image:
        min_val = np.min(image)
        max_val = np.max(image)
        image = np.floor((image - min_val) / (max_val - min_val) * 255).astype(np.uint8)

    if len(image.shape) == 2:
        # make it artificially rgb
        image = np.repeat(image[:, :, None], 3, axis=-1)

    return image


class LocalImage(BaseImage):
    """Class for a single image"""

    def __init__(self, content, frame=None):
        self.content = content
        self.frame = frame

    @property
    def raw(self):
        return self.content

    @property
    def num_channels(self):
        if len(self.raw.shape) == 2:
            # only width and height -> 1 channel
            return 1
        else:
            # multiple channels -> channels are specified at the end
            return self.raw.shape[-1]

    def get_channel(self, channel: int):
        assert channel < self.num_channels

        if self.num_channels == 1 and len(self.raw.shape) == 2:
            return self.raw
        else:
            return self.raw[..., channel]

    def __getitem__(self, item):
        return self.raw[item]


class LocalImageSource(ImageSequenceSource):
    """Source for a single image only"""

    def __init__(self, image: LocalImage):
        self.image = image

    def __get_image(self):
        return self.image

    def __iter__(self):
        yield self.__get_image()

    def get_frame(self, frame: int):
        assert frame == 0, f"We only have a single frame, but frame={frame}"

        return self.__get_image()

    @property
    def num_channels(self) -> int:
        return self.__get_image().num_channels

    @property
    def num_frames(self) -> int:
        return 1

    def __len__(self):
        return 1

    @staticmethod
    def from_file(file_path: str, normalize_image=True):
        image = LocalImage(prepare_image(cv2.imread(file_path), normalize_image))

        return LocalImageSource(image)

    @staticmethod
    def from_array(array):
        image = LocalImage(array)

        return LocalImageSource(image)


class InMemorySequenceSource(ImageSequenceSource):
    """Image sequence for an in memory image stack"""

    def __init__(self, image_stack):
        self.image_stack = image_stack

    def get_frame(self, frame: int) -> BaseImage:
        assert frame < len(self.image_stack)

        return LocalImage(self.image_stack[frame])

    def __len__(self):
        return len(self.image_stack)

    def __iter__(self):
        for i in range(len(self)):
            yield self.get_frame(i)

    @property
    def num_channels(self) -> int:
        return self.get_frame(0).num_channels


class LocalSequenceSource(ImageSequenceSource):
    """Image sequence source for files in the local file system (e.g. a tif)."""

    def __init__(
        self, tif_file: str, normalize_image=True, luts=None, channel_index: int = 0
    ):
        """Create a new local image source

        Args:
            tif_file (str): path to the image file
            normalize_image (bool, optional): Normalizes the image pixels t0 [0, 255]. Defaults to True.
            luts: (List, optional): List of lut functions applied to the channels
            channel_index (int, optional): index in image of the channel. For example, for H,W,C dims where C is channel we should have a 2.
        """
        self.filename = tif_file
        self.normalize_image = normalize_image
        self.luts = luts
        self.channel_index = channel_index

    def __iter__(self):
        images = tifffile.imread(self.filename)

        for image in images:
            if self.luts is not None:
                if len(image.shape) == 2:
                    # just a single channel
                    num_image_channels = 1
                else:
                    num_image_channels = image.shape[self.channel_index]

                assert (
                    len(self.luts) == num_image_channels
                ), f"We need a LUTs function for every channel! We have {num_image_channels} channels but only {len(self.luts)} LUTs!"
                # apply luts to image
                if len(image.shape) == 2:
                    # we only have one channel
                    image = self.luts[0](image)
                elif len(image.shape) == 3:
                    # we have N channels (at the front)
                    for channel in range(image.shape[self.channel_index]):
                        image[channel] = self.luts[channel](
                            image.take(channel, axis=self.channel_index)
                        )

            image = prepare_image(image, self.normalize_image)

            yield LocalImage(image)

    def get_frame(self, frame: int) -> BaseImage:
        # TODO: this is super slow access for indiviudal images
        images = tifffile.imread(self.filename)
        assert frame < len(images)

        return LocalImage(prepare_image(images[frame]))

    @property
    def num_channels(self) -> int:
        return self.get_frame(0).num_channels

    def slice(self, start, end):
        images = tifffile.imread(self.filename)

        for image in images[start:end]:
            # normalize image space
            if self.normalize_image:
                min_val = np.min(image)
                max_val = np.max(image)
                image = np.floor((image - min_val) / (max_val - min_val) * 255).astype(
                    np.uint8
                )

            if len(image.shape) > 2:
                # select only the first channel
                image = image[0]

            if len(image.shape) == 2:
                # make it artificially rgb
                image = np.repeat(image[:, :, None], 3, axis=-1)

            yield LocalImage(image)


class ImageJRoISource(RoISource):
    """Source fro ImageJ RoI file format"""

    def __init__(self, filename, range=None):
        self.overlay = RoiStorer.load(filename)
        self.range = range

    def __iter__(self):
        return self.overlay.timeIterator(frame_range=self.range)

    def __len__(self) -> int:
        if self.range:
            min(len(self.overlay), len(self.range))
        return len(self.overlay)


class RoiStorer:
    """
    Stores and loads overlay results in the roi format (readable by ImageJ)
    """

    @staticmethod
    def store(overlay: Overlay, filename: str, append=False):
        """
        Stores overlay results in the roi format (readable by fiji)

        overlay: the overlay to store
        filename: filename of the roi collection (e.g. rois.zip)
        append: appends the rois if the file already exists
        """

        # generate imagej rois from the overlay
        rois = [
            roifile.ImagejRoi.frompoints(contour.coordinates, t=contour.frame)
            for contour in overlay
        ]

        if not append and osp.isfile(filename):
            os.remove(filename)

        # write them to file
        roifile.roiwrite(filename, rois)

    @staticmethod
    def load(filename: str):
        # read the imagej rois from file
        rois = roifile.roiread(filename)

        id = -1
        # convert them into contours (recover time position)
        contours = [
            Contour(roi.coordinates(), -1.0, roi.position - 1, id=id) for roi in rois
        ]

        # return the overlay
        return Overlay(contours)



"""Module for testing visualization module"""

import unittest

import numpy as np

from acia.viz import draw_scale_bar, draw_time


class TestOverlays(unittest.TestCase):
    """Testing scalebar and time overlay"""

    def test_time_overlay(self):
        """Simply draw time overlay on empty image"""

        images = np.zeros((10, 200, 200, 3), dtype=np.uint8)

        for image in draw_time(images, (100, 100), "5 minutes"):
            self.assertTrue(isinstance(image, np.ndarray))
            np.testing.assert_array_equal(image.shape, (200, 200, 3))

    def test_time_overlay_with_bg(self):
        """Simply draw time overlay on empty image with background"""

        images = np.zeros((10, 200, 200, 3), dtype=np.uint8)

        for image in draw_time(
            images, (100, 100), "5 minutes", background_color=(0, 0, 0)
        ):
            self.assertTrue(isinstance(image, np.ndarray))
            np.testing.assert_array_equal(image.shape, (200, 200, 3))

    def test_scale_bar_overlay(self):
        """Simply draw scale bar on empty image"""

        images = np.zeros((10, 200, 200, 3), dtype=np.uint8)

        for image in draw_scale_bar(
            images,
            xy_position=(100, 100),
            size_of_pixel="0.07 micrometer/pixel",
            bar_width="5 micrometer",
            bar_height="0.25 micrometer",
        ):
            self.assertTrue(isinstance(image, np.ndarray))
            np.testing.assert_array_equal(image.shape, (200, 200, 3))

    def test_scale_bar_overlay_with_background(self):
        """Simply draw scale bar on empty image with a background color"""

        images = np.zeros((10, 200, 200, 3), dtype=np.uint8)

        for image in draw_scale_bar(
            images,
            xy_position=(100, 100),
            size_of_pixel="0.07 micrometer/pixel",
            bar_width="5 micrometer",
            bar_height="0.25 micrometer",
            background_color=(255, 0, 0),
        ):
            self.assertTrue(isinstance(image, np.ndarray))
            np.testing.assert_array_equal(image.shape, (200, 200, 3))


if __name__ == "__main__":
    unittest.main()



""" Functions for different segmentation formats"""


import json
from pathlib import Path

import numpy as np
from tifffile import imread

from acia.base import Contour, Instance, Overlay
from acia.utils import multi_mask_to_polygons


def parse_simple_segmentation(file_content: str) -> Overlay:
    """Parse simple segmentation format from string (json)

    Args:
        file_content (str): simple segmentation file content as string

    Returns:
        Overlay: the overlay representation of the segmentation
    """

    file_data = json.loads(file_content)

    contours = []

    for frame in file_data:
        frame_id = frame["frame"]
        for det in frame["detections"]:
            contours.append(
                Contour(det["contour"], -1.0, frame_id, det["id"], det["label"])
            )

    return Overlay(contours)


def gen_simple_segmentation(overlay: Overlay) -> str:
    """Create a simple segmentation string from an Overlay

    Args:
        overlay (Overlay): the overlay to store

    Returns:
        str: string containing the stringified simple segmentation json format
    """

    frame_packages = []

    # loop over frames
    for frame_overlay in overlay.timeIterator():
        if len(frame_overlay) == 0:
            continue

        det_objects = []
        frame_id = -1

        # loop over all contours in a frame
        for cont in frame_overlay:

            # transform coordinates to list (otherwise not json serializable)
            coordinates = cont.coordinates
            if isinstance(coordinates, np.ndarray):
                coordinates = coordinates.tolist()

            # create detection object
            det_objects.append(
                dict(
                    label=cont.label,
                    contour=coordinates,
                    id=cont.id,
                )
            )

            frame_id = cont.frame

        # create the frame package
        frame_package = dict(
            frame=frame_id,
            detections=det_objects,
        )
        frame_packages.append(frame_package)

    # serialize into json format
    return json.dumps(frame_packages, separators=(",", ":"))


def load_ctc_segmentation(segmentation_path: Path) -> Overlay:
    segmentation_path = Path(segmentation_path)

    segm_mask_files = sorted(segmentation_path.glob("*.tif"))

    overlay = Overlay([], frames=list(range(len(segm_mask_files))))

    c_id = 0

    for frame_id, segm_file in enumerate(segm_mask_files):
        polygons = multi_mask_to_polygons(imread(segm_file))

        for _, poly in polygons:
            points = np.array(poly.exterior.coords.xy)
            overlay.add_contour(Contour(points, -1, frame_id, c_id, "cell"))
            c_id += 1

    return overlay


def load_ctc_segmentation_native(segmentation_path: Path) -> Overlay:
    """Fast loading of CTC segmentation masks into an Overlay

    Args:
        segmentation_path (Path): Path to the folder containing all the *.tif masks

    Returns:
        Overlay: Overlay containing all masks
    """

    # List all the segmentation masks
    segmentation_path = Path(segmentation_path)
    segm_mask_files = sorted(segmentation_path.glob("*.tif"))

    overlay = Overlay([], frames=list(range(len(segm_mask_files))))

    # unique id for instances
    uid = 1

    # Iterate all the mask files
    for frame_id, segm_file in enumerate(segm_mask_files):

        # read the tif
        mask = imread(segm_file)

        # Find all cell labels (except 0)
        labels = np.unique(mask)[1:]

        # for every label create an instance and add it to the contour
        for label in labels:
            instance = Instance(mask=mask, frame=frame_id, label=label)
            instance.id = uid

            overlay.add_contour(instance)
            uid += 1

    return overlay



"""Classes for OMERO storage interaction"""

from __future__ import annotations

import logging
from pathlib import Path

import numpy as np
import omero
import tqdm.auto as tqdm
from omero.gateway import BlitzGateway
from omero.model import LengthI

from acia import ureg
from acia.base import BaseImage, Contour, ImageSequenceSource, Overlay, RoISource
from acia.segm.local import LocalImage
from acia.segm.omero.shapeUtils import make_coordinates
from acia.segm.utils import compute_indices

from .shapeUtils import create_polygon


# We have a helper function for creating an ROI and linking it to new shapes
def create_roi(updateService, img, shapes):
    """
    Helper function to create the roi object
    updateService: omero update Service
    img: omero image object (not the id)
    shapes: list of omero.model shapes

    returns: RoI object
    """
    # create an ROI, link it to Image
    roi = omero.model.RoiI()
    # use the omero.model.ImageI that underlies the 'image' wrapper
    roi.setImage(img._obj)
    for shape in shapes:
        roi.addShape(shape)
    # Save the ROI (saves any linked shapes too)
    return updateService.saveAndReturnObject(roi)


# We have a helper function for creating an ROI and linking it to new shapes
def create_roi_fast(updateService, img, shapes):
    """
    Helper function to create the roi object (without waiting)
    updateService: omero update Service
    img: omero image object (not the id)
    shapes: list of omero.model shapes
    """
    # create an ROI, link it to Image
    roi = omero.model.RoiI()
    # use the omero.model.ImageI that underlies the 'image' wrapper
    roi.setImage(img._obj)
    for shape in shapes:
        roi.addShape(shape)
    # Save the ROI (saves any linked shapes too)
    updateService.saveObject(roi)


class OmeroRoIStorer:
    """
    Stores and loads overlay results in the roi format (readable by ImageJ)
    """

    @staticmethod
    def storeWithConn(overlay, imageId: int, conn, force=False):
        # retrieve omero objects
        updateService = conn.getUpdateService()
        image = conn.getObject("Image", imageId)

        userId = conn.getUser().getId()
        imageOwnerId = image.getOwner().getId()

        if not force and userId != imageOwnerId:
            raise ValueError(
                "You try to write to non-owned data. Enable 'force' option if you are sure to do that."
            )

        # OmeroRoIStorer.clear(imageId=imageId, username=username, password=password, serverUrl=serverUrl, port=port, secure=secure)

        # size_t = image.getSizeT()
        size_z = image.getSizeZ()

        # this is a linearized overlay
        logging.info(
            "Using Linearized overlay: [t_1: z_0, z_1, ... z_Z, t_2: ,...] Use t and z"
        )
        shapes = [
            create_polygon(
                cont.coordinates,
                z=cont.frame % size_z,
                t=np.floor(cont.frame / size_z),
                description=f"Score: {cont.score:.2f}",
            )
            for cont in overlay
        ]

        for shape in tqdm.tqdm(shapes):
            create_roi_fast(updateService, image, [shape])

        logging.info(
            "Stored overlay with %d rois for image '%s'", len(overlay), image.getName()
        )

    @staticmethod
    def store(
        overlay: Overlay,
        imageId: int,
        username: str,
        password: str,
        serverUrl: str,
        port=4064,
        secure=True,
        force=False,
        conn=None,
    ):
        """
        Stores overlay results in OMERO. Uses existing connection if available.

        overlay: the overlay to store
        imageId: omero id of the image sequence
        username: omero username
        password: omero password
        serverUrl: omero web address
        port: omero port (default: 4064)
        conn: existing OMERO connection or None. (default: None)
        """

        with BlitzConn(username, password, serverUrl, port, secure, conn) as omero_conn:
            OmeroRoIStorer.storeWithConn(overlay, imageId, omero_conn, force)

    @staticmethod
    def load(
        imageId: int,
        username: str,
        password: str,
        serverUrl: str,
        port=4064,
        secure=True,
        roiId=None,
        conn=None,
    ) -> Overlay:
        """
        Loads overlay from omero. Only considers polygons.

        imageId: omero id of the image sequence
        username: omero username
        password: omero password
        serverUrl: omero web address
        port: omero port (default: 4064)
        """
        overlay = Overlay([])
        # open connection to omero
        # with BlitzGateway(
        #    username, password, host=serverUrl, port=port, secure=secure
        # ) as conn:
        with BlitzConn(username, password, serverUrl, port, secure, conn) as omero_conn:
            # get the roi service
            roi_service = omero_conn.getRoiService()
            result = roi_service.findByImage(imageId, None)

            image = omero_conn.getObject("Image", imageId)

            # size_t = image.getSizeT()
            size_z = image.getSizeZ()

            # loop rois
            for roi in result.rois:
                if (roiId is not None) and roi.getId() != roiId:
                    # if the roiId is specified, check whether we have the right one
                    continue

                # loop shapes inside roi
                for s in roi.copyShapes():
                    if isinstance(s, omero.model.PolygonI):
                        # extract important information
                        t = s.getTheT().getValue()
                        points = make_coordinates(s.getPoints().getValue())
                        score = -1.0

                        if size_z > 1:
                            t = t * size_z + s.getTheZ().getValue()

                        id = s.getId().getValue()

                        label = s.getTextValue().getValue()

                        # add contour element to overlay
                        cont = Contour(points, score, t, id=id, label=label)
                        overlay.add_contour(cont)

        # return the overlay
        return overlay

    @staticmethod
    def clear(
        imageId: int,
        username: str = None,
        password: str = None,
        serverUrl: str = None,
        port=4064,
        secure=True,
        conn=None,
    ):
        # open connection to omero
        with BlitzConn(
            username=username,
            password=password,
            serverUrl=serverUrl,
            port=port,
            secure=secure,
            conn=conn,
        ) as omero_conn:
            # get the roi service
            roi_service = omero_conn.getRoiService()
            updateService = omero_conn.getUpdateService()
            result = roi_service.findByImage(imageId, None)

            print(f"Deleting {len(result.rois)} rois...")

            for roi in result.rois:
                shapes = roi.copyShapes()
                if len(shapes) > 1:
                    for s in roi.copyShapes():
                        roi.removeShape(s)
                    roi = updateService.saveAndReturnObject(roi)

            # delete all RoIs in the image
            if len(result.rois) > 0:
                omero_conn.deleteObjects(
                    "Roi",
                    [roi.getId().getValue() for roi in result.rois],
                    deleteAnns=True,
                    deleteChildren=True,
                    wait=True,
                )


class IngoreWithWrapper:
    """Wrapper to ignore context and do not exit after leaving. Is used for not starting/ending omero connections all the time!"""

    def __init__(self, object):
        self.object = object

    def __getattr__(self, attr):
        return getattr(self.object, attr)

    def __enter__(self):
        return self.object

    def __exit__(self, type, value, traceback):
        pass


class BlitzConn:
    """
    Encapsulates standard omero behavior
    """

    def __init__(
        self,
        username,
        password,
        serverUrl,
        port=4064,
        secure=True,
        conn=None,
        readonly=True,
    ):

        assert username is not None or conn is not None, "Please provide a username"
        assert password is not None or conn is not None, "Please provide a password"
        assert (
            serverUrl is not None or conn is not None
        ), "Please provide a OMERO server"

        self.username = username
        self.password = password
        self.serverUrl = serverUrl
        self.port = port
        self.secure = secure

        self.conn = conn
        self.readonly = readonly

    def make_connection(self):
        # try to keep connection alive
        if self.conn is not None and self.conn.keepAlive():
            return IngoreWithWrapper(self.conn)
        else:
            # return a new connection
            conn = BlitzGateway(
                self.username,
                self.password,
                host=self.serverUrl,
                port=self.port,
                secure=self.secure,
            )

            # establish connection
            connect_res = conn.connect()

            # check connection
            if connect_res is False:
                raise ConnectionError(
                    "Connection to OMERO failed! Please check your OMERO settings (username, password, omero host, ...)!"
                )

            # if readonly -> set group so that we can immediately access all data
            if self.readonly:
                conn.SERVICE_OPTS.setOmeroGroup("-1")
            # store the connection
            self.conn = conn
            return IngoreWithWrapper(self.conn)

    def __enter__(self):
        return self.make_connection()

    def __exit__(self, type, value, traceback):
        pass

    def __del__(self):
        # make sure the connection is always closed
        if self.conn:
            self.conn.close()
            self.conn = None


class OmeroSource(BlitzConn):
    """
    Base Class for omero image information. Bundles functionality for image and RoIs.
    """

    def __init__(
        self,
        imageId: float,
        username: str = None,
        password: str = None,
        serverUrl: str = None,
        port=4064,
        secure=True,
        conn=None,
        readonly=True,
    ):
        """
        Args:
            imageId (float): omero image id
            username (str, optional): omero username. Not needed when conn is provided. Defaults to None.
            password (str, optional): omero password. Not needed when conn is provided. Defaults to None.
            serverUrl (str, optional): omero url. Not needed when conn is provided. Defaults to None.
            port (int, optional): omero port. Not needed when conn is provided. Defaults to 4064.
            secure (bool, optional): Whether to choose secure connection. Defaults to True.
            conn ([type], optional): Existing omero connection. Defaults to None.
        """
        BlitzConn.__init__(
            self,
            username=username,
            password=password,
            serverUrl=serverUrl,
            port=port,
            secure=secure,
            conn=conn,
            readonly=readonly,
        )

        self.imageId = imageId

    @property
    def rawPixelSize(self) -> tuple[LengthI, LengthI]:
        """Return the pixel size in omero objects

        Returns:
            Tuple[LengthI,LengthI]: x and y pixel size in omero objects
        """
        with self.make_connection() as conn:
            image = conn.getObject("Image", self.imageId)

            size_x_obj = image.getPixelSizeX(units="MICROMETER")
            size_y_obj = image.getPixelSizeY(units="MICROMETER")

            return size_x_obj, size_y_obj

    @property
    def pixelSize(self) -> tuple[float, float]:
        """Return the pixel size in micron

        Returns:
            Tuple[float,float]: x and y pixel size in micron
        """
        size_x_obj, size_y_obj = self.rawPixelSize

        return size_x_obj.getValue(), size_y_obj.getValue()

    @property
    def pixel_size(self) -> tuple[ureg.Quantity, ureg.Quantity]:
        """Return the pixel size in micrometer

        Returns:
            Tuple[float,float]: x and y pixel size in micrometer
        """
        size_x_obj, size_y_obj = self.rawPixelSize

        unit = "micrometer".upper()

        return (
            omero.model.LengthI(size_x_obj, unit).getValue() * ureg.micrometer,
            omero.model.LengthI(size_y_obj, unit).getValue() * ureg.micrometer,
        )

    def printPixelSize(self, unit="MICROMETER"):
        """Output pixel sizes

        Args:
            unit (str, optional): Name of the unit. Defaults to "MICROMETER".
        """
        # get raw
        size_x_obj, size_y_obj = self.rawPixelSize

        # convert to correct unit
        size_x_obj = omero.model.LengthI(size_x_obj, unit)
        size_y_obj = omero.model.LengthI(size_y_obj, unit)

        # output pixel sizes
        print(f" Pixel Size X: {size_x_obj.getValue()} ({size_x_obj.getSymbol()})")
        print(f" Pixel Size Y: {size_y_obj.getValue()} ({size_y_obj.getSymbol()})")


class OmeroSequenceSource(ImageSequenceSource, OmeroSource):
    """
    Uses omero server as a source for images
    """

    def __init__(
        self,
        imageId: int,
        username: str = None,
        password: str = None,
        serverUrl: str = None,
        port=4064,
        channels=None,
        z=0,
        imageQuality=1.0,
        secure=True,
        colorList=None,
        range=None,
        conn=None,
    ):
        """
        imageId: id of the image sequence
        username: omero username
        password: omero password
        serverUrl: omero server url
        port: omero port
        channels: list of image channels to activate (e.g. include fluorescence channels)
        z: focus plane
        imageQuality: quality of the rendered images (1.0=no compression, 0.0=super compression)
        base_channel: id of the phase contrast channel (visualized over all rgb channels)
        """

        OmeroSource.__init__(
            self,
            imageId=imageId,
            username=username,
            password=password,
            serverUrl=serverUrl,
            port=port,
            secure=secure,
            conn=conn,
            readonly=True,
        )

        if channels is None:
            channels = [1]
        if colorList is None:
            colorList = ["FFFFFF", None, None]

        self.imageId = imageId
        self.channels = channels
        self.z = z
        self.imageQuality = imageQuality
        self.colorList = colorList
        self.range = range

        if self.range is not None:
            # we make it a list
            self.range = list(self.range)

            # we have a look that it is not tool long
            if np.max(self.range) > len(self):
                logging.warning(
                    "Range exceeds number of images! Truncate to %d images", len(self)
                )
                np_range = np.array(self.range)
                self.range = np_range[np_range < len(self)]

        assert len(self.channels) <= len(
            self.colorList
        ), f"you must specify a color for every channel! You have {len(self.channels)} channels ({self.channels}) but only {len(self.colorList)} color(s) ({self.colorList}). Please update your colorList!"

    def imageName(self) -> str:
        """
        returns the name of the image
        """
        with self.make_connection() as conn:
            return conn.getObject("Image", self.imageId).getName()

    def datasetName(self) -> str:
        """
        returns the name of the dataset
        """
        with self.make_connection() as conn:
            return conn.getObject("Image", self.imageId).getParent().getName()

    def projectName(self) -> str:
        """
        returns the name of the associated project
        """
        with self.make_connection() as conn:
            return conn.getObject("Image", self.imageId).getProject().getName()

    def __get_omero_image(self):
        # open the connection
        with self.make_connection() as conn:
            return conn.getObject("Image", self.imageId)

    def __get_image(self, frame: int) -> BaseImage:
        # get the specified image
        image = self.__get_omero_image()

        size_t = image.getSizeT()
        size_z = image.getSizeZ()

        t, z = compute_indices(frame, size_t, size_z)

        # perform rendering
        image.setColorRenderingModel()
        image.setActiveChannels(self.channels, colors=self.colorList)
        rendered_image = image.renderImage(z, t, compression=self.imageQuality)

        # return local image
        return LocalImage(np.asarray(rendered_image, dtype=np.uint8), frame=frame)

    def __iter__(self):
        for frame in self.frame_list:
            if self.range is not None and (frame not in self.range):
                continue
            yield self.get_frame(frame)

    def get_frame(self, frame: int):
        return self.__get_image(frame)

    @property
    def num_channels(self) -> int:
        return len(self.channels)

    @property
    def num_frames(self) -> int:
        return len(self)

    @property
    def frame_list(self) -> list[int]:
        if self.range is not None:
            return list(self.range)
        else:
            return list(range(len(self)))

    def __len__(self):
        with self.make_connection() as conn:
            image = conn.getObject("Image", self.imageId)
            if self.range is not None:
                return min(image.getSizeT() * image.getSizeZ(), len(self.range))
            return int(image.getSizeT() * image.getSizeZ())


class OmeroRawSource(ImageSequenceSource, OmeroSource):
    """Raw OMERO source: Allows to easily access raw that is, e.g. 16-bit data of your OMERO images"""

    def __init__(
        self,
        imageId: int,
        username: str = None,
        password: str = None,
        serverUrl: str = None,
        port=4064,
        secure=True,
        conn=None,
        channels=None,
    ):
        OmeroSource.__init__(
            self,
            imageId=imageId,
            username=username,
            password=password,
            serverUrl=serverUrl,
            port=port,
            secure=secure,
            conn=conn,
        )

        if channels is None:
            channels = [0]

        self.channels = channels

    def __len__(self):
        with self.make_connection() as conn:
            image = conn.getObject("Image", self.imageId)

            size_z = image.getSizeZ()
            size_t = image.getSizeT()

            return size_z * size_t

    def get_frame(self, frame):
        with self.make_connection() as conn:
            # Use the pixelswrapper to retrieve the plane as
            # a 2D numpy array see [https://github.com/scipy/scipy]
            #
            # Numpy array can be used for various analysis routines
            #
            image = conn.getObject("Image", self.imageId)
            size_z = image.getSizeZ()
            size_t = image.getSizeT()

            pixels = image.getPrimaryPixels()

            t, z = compute_indices(frame, size_t, size_z)

            planes = []
            for channel in self.channels:
                c = channel

                planes.append(pixels.getPlane(z, c, t))
            return LocalImage(np.stack(planes, axis=-1))

    def __iter__(self):
        with self.make_connection() as conn:
            # Use the pixelswrapper to retrieve the plane as
            # a 2D numpy array see [https://github.com/scipy/scipy]
            #
            # Numpy array can be used for various analysis routines
            #
            image = conn.getObject("Image", self.imageId)
            size_z = image.getSizeZ()
            size_c = image.getSizeC()
            size_t = image.getSizeT()

            pixels = image.getPrimaryPixels()

            for t in range(size_t):
                for z in range(size_z):
                    planes = []
                    for channel in self.channels:
                        assert channel < size_c, "Please specify a valid channel"
                        c = channel

                        planes.append(pixels.getPlane(z, c, t))
                    yield LocalImage(np.stack(planes, axis=-1))

    @property
    def num_channels(self) -> int:
        return len(self.channels)


class OmeroRoISource(OmeroSource, RoISource):
    """Source for OMERO RoIs beloging to an OMERO image sequence"""

    def __init__(
        self,
        imageId: int,
        username: str,
        password: str,
        serverUrl: str,
        port=4064,
        secure=True,
        roiSelector=lambda rois: [rois[0]],
        range=None,
        scale=None,
        conn=None,
    ):
        OmeroSource.__init__(
            self,
            imageId=imageId,
            username=username,
            password=password,
            serverUrl=serverUrl,
            port=port,
            secure=secure,
            conn=conn,
        )

        self.imageId = imageId

        self.roiSelector = roiSelector
        self.range = range
        self.scale = scale
        if self.scale:
            # 1 pixel has the size of the returned value. To move to correct domain use that size as scale factor
            self.scaleFactor = omero.model.LengthI(
                self.rawPixelSize[0], self.scale
            ).getValue()

        self.overlay = None

    def __iter__(self):
        # return overlay iterator over time
        return self.get_overlay().timeIterator(frame_range=self.range)

    def get_overlay(self):
        if self.overlay is None:
            # compose an overlay from the rois
            self.overlay = OmeroRoIStorer.load(
                self.imageId,
                username=self.username,
                password=self.password,
                serverUrl=self.serverUrl,
                port=self.port,
                secure=self.secure,
                conn=self.make_connection(),
            )

            if self.scale:
                self.overlay.scale(self.scaleFactor)

        return self.overlay

    def __len__(self) -> int:
        with self.make_connection() as conn:
            image = conn.getObject("Image", self.imageId)
            if self.range is not None:
                return min(image.getSizeT() * image.getSizeZ(), len(self.range))
            return image.getSizeT() * image.getSizeZ()


def upload_file(
    omero_type: str,
    omero_id: int,
    file_path: Path,
    conn: BlitzGateway,
    mime_type="text/plain",
    namespace: str = None,
):
    """Upload a file attachement for an OMERO object (Image, Datase, Project)

    Args:
        omero_type (str): Image, Dataset or Project
        omero_id (int): id of the omero object
        file_path (Path): file path of the file to upload
        conn (BlitzGateway): connection to OMERO
        mime_type (str, optional): mime type of the uploaded data. Defaults to "text/plain".
        namespace(str, optional): OMERO namespace of the uploaded file

    Returns:
        _type_: created annotation object
    """

    omero_object = conn.getObject(omero_type, omero_id)

    # create the original file and file annotation (uploads the file etc.)
    file_ann = conn.createFileAnnfromLocalFile(
        file_path, mimetype=mime_type, ns=namespace, desc=None
    )

    logging.info(
        "Attaching FileAnnotation to OMERO object: File ID: %d, %s Size: %d",
        file_ann.getId(),
        file_ann.getFile().getName(),
        file_ann.getFile().getSize(),
    )

    # link it to dataset.
    omero_object.linkAnnotation(file_ann)

    # return OMERO annotation
    return file_ann


def download_file_from_object(
    omero_type: str,
    omero_id: int,
    file_name: str,
    output_path: Path,
    conn,
    append_filename=False,
):
    """Download a file annotation with a given name from an OMERO object

    Args:
        omero_type (str): Image, Dataset or Project
        omero_id (int): unique id of the OMERO object
        file_name (str): the name of the file attachment to download
        output_path (Path): output path to save the file
        conn (_type_): OMERO connection
        append_filename (bool, optional): If true the filename is appended to the output path. Defaults to False.

    Raises:
        ValueError: _description_
    """
    if not isinstance(output_path, Path):
        output_path = Path(output_path)

    anns = list_file_annotations(omero_type=omero_type, omero_id=omero_id, conn=conn)

    anns = list(filter(lambda ann: ann.getFile().getName() == file_name, anns))

    if len(anns) != 1:
        raise ValueError(f"Annotation count is not 1 but: {len(anns)}")

    download_file_from_annotation(anns[0], output_path, append_filename)


def download_file_from_annotation(ann, output_path: Path, append_filename=False):
    """Download OMERO file attachment

    Args:
        ann (_type_): OMERO annotation object
        output_path (Path): path to write the file to
        append_filename (bool, optional): If True treats output_path as a folder and appends the OMERO file name. Defaults to False.
    """
    assert isinstance(ann, omero.gateway.FileAnnotationWrapper)

    if not isinstance(output_path, Path):
        output_path = Path(output_path)

    logging.info(
        "File ID: %d %s Size: %d",
        ann.getFile().getId(),
        ann.getFile().getName(),
        ann.getFile().getSize(),
    )

    file_path = output_path

    if append_filename:
        file_path = file_path / ann.getFile().getName()

    with open(str(file_path), "wb") as f:
        logging.info("\nDownloading file to %s...", file_path)
        for chunk in ann.getFileInChunks():
            f.write(chunk)
    logging.info("File downloaded!")


def delete_file_annotation(annotation_id: int, conn: BlitzGateway):
    """Delete an OMERO file attachement

    Args:
        annotation_id (int): id of the annotation object
        conn (BlitzGateway): OMERO connection
    """
    # ann_obj = conn.getObject('Annotation', annotation_id)
    conn.deleteObjects("Annotation", [annotation_id], wait=True)


def list_file_annotations(omero_type: str, omero_id: int, conn: BlitzGateway):
    """List all file annotations of an OMERO object

    Args:
        omero_type (str): Image, Dataset or Project OMERO object type
        omero_id (int): unique OMERO id of the object
        conn (BlitzGateway): OMERO connection

    Returns:
        _type_: List of OMERO annotation objects for files
    """
    # get OMERO object
    omero_object = conn.getObject(omero_type, omero_id)

    annotations = []

    for ann in omero_object.listAnnotations():
        if isinstance(ann, omero.gateway.FileAnnotationWrapper):
            # add file annotations
            annotations.append(ann)

    return annotations


def replace_file_annotation(
    omero_type: str, omero_id: int, file_path: Path, conn, mime_type: str = "text/plain"
):
    """Replace an OMERO file by a new version.

    Deletes existing file attachements on OMERO with the same name and uploads the new file.

    Args:
        omero_type (str): Image, Dataset or Project OMERO object type
        omero_id (int): unique OMERO id of the object
        file_path (Path): path of the file to upload
        conn (_type_): OMERO connection
        mime_type (str, optional): MIME type of the data file to upload. Defaults to "text/plain".

    Returns:
        _type_: new annotation object of the uploaded file
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)

    # only get annotations with the same filename
    anns = filter(
        lambda ann: ann.getFile().getName() == file_path.name,
        list_file_annotations(omero_type, omero_id, conn),
    )

    # upload the new file version first
    new_ann = upload_file(omero_type, omero_id, file_path, conn, mime_type=mime_type)

    # delete every other annotation with the same name
    for ann in anns:
        delete_file_annotation(ann.getId(), conn)

    return new_ann


def print_file_annotations(omero_type: str, omero_id: int, conn: BlitzGateway):
    """List all file annotations

    Args:
        omero_type (str): Image, Dataset or Project OMERO object type
        omero_id (int): unique OMERO id of the object
        conn (BlitzGateway): OMERO connection
    """
    for ann in list_file_annotations(omero_type, omero_id, conn):
        print(
            "File ID:",
            ann.getFile().getId(),
            ann.getFile().getName(),
            "Size:",
            ann.getFile().getSize(),
            "Namespace:",
            ann.getNs(),
        )



"""Functionality for creating outputs (export) the segmentation information"""

from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Iterable, Literal

import cv2
import numpy as np
import tqdm.auto as tqdm

from acia.base import BaseImage, ImageRoISource, ImageSequenceSource, Overlay
from acia.utils import ScaleBar
from acia.viz import VideoExporter2


class DatasetExporter:
    """Base class for dataset exporters"""

    def __init__(self):
        self.sources = []

    def add(self, item: ImageRoISource | list[ImageRoISource]):
        if isinstance(item, Iterable):
            # iterable
            self.sources += item
        else:
            # not iterable
            self.sources.append(item)


class MMSegmentationDataset(DatasetExporter):
    """MMSegmentation dataset exporter"""

    def __init__(
        self,
        labels=None,
        label_coverter=lambda x: x,
    ):
        super().__init__()

        if labels is None:
            labels = ["Stem", "ThickRoot", "MediumRoot", "ThinRoot"]

        self.labels = labels
        self.label_converter = label_coverter

    def write(self, base_folder: str | Path = "data", mode="train"):

        img_path = Path(base_folder).absolute() / "img_dir" / mode
        ann_path = Path(base_folder).absolute() / "ann_dir" / mode

        img_path.mkdir(parents=True, exist_ok=True)
        ann_path.mkdir(parents=True, exist_ok=True)

        for input_index, image_roi_source in enumerate(tqdm.tqdm(self.sources)):
            for image_index, (image, rois) in enumerate(image_roi_source):
                if len(rois) == 0:
                    print("Skip")
                    continue

                # save image file
                image_file_path = img_path / f"{input_index:03d}_{image_index:03d}.png"
                mask_file_path = ann_path / f"{input_index:03d}_{image_index:03d}.png"
                cv2.imwrite(
                    str(image_file_path), cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
                )

                height, width = image.shape[:2]

                image_mask = np.zeros((height, width), dtype=np.uint8)

                # save roi masks
                for i, label in enumerate(self.labels):
                    label_value = i + 1
                    rois_for_label = filter(
                        partial(
                            lambda r, label: self.label_converter(r.label) == label,
                            label=label,
                        ),
                        rois,
                    )

                    label_mask = np.zeros((height, width), dtype=np.uint8)

                    for roi in rois_for_label:
                        roi_mask = roi.toMask(
                            height,
                            width,
                            fillValue=label_value,
                            outlineValue=label_value,
                        )

                        label_mask |= roi_mask

                    kernel = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]], dtype=np.uint8)

                    dil_label_mask = cv2.dilate(label_mask, kernel)
                    ero_label_mask = cv2.erode(dil_label_mask, kernel)

                    label_mask = ero_label_mask

                    image_mask[label_mask > 0] = (
                        label_mask.astype(np.uint8) * label_value
                    )[label_mask > 0]

                cv2.imwrite(str(mask_file_path), image_mask)


def no_crop(frame: int, _: Overlay):
    return (slice(0, frame.shape[0]), slice(0, frame.shape[1]))


def __video_export_from_str(
    filename: str, codec: Literal["vp09", "mjpg", "h264", "h265"], framerate: int
) -> VideoExporter2:
    """Create video exporter from string

    Args:
        filename (str): video filename
        codec (Literal[&quot;vp09&quot;, &quot;mjpg&quot;, &quot;h264&quot;, &quot;h265&quot;]): codec string
        framerate (int): framerate (in fps)

    Raises:
        ValueError: If codec is not found

    Returns:
        VideoExporter2: generate appropriate video exporter
    """
    if codec == "vp09":
        ve = VideoExporter2.default_vp9(filename=filename, framerate=framerate)
    elif codec == "h264":
        ve = VideoExporter2.default_h264(filename=filename, framerate=framerate)
    elif codec == "h265":
        ve = VideoExporter2.default_h265(filename=filename, framerate=framerate)
    elif codec == "mjpg":
        ve = VideoExporter2.default_mjpg(filename=filename, framerate=framerate)
    else:
        raise ValueError(f"Unknown/Unsupported codec: {codec}")

    return ve


def renderVideo(
    imageSource: ImageSequenceSource,
    roiSource=None,
    filename="output.mp4",
    framerate=3,
    codec: Literal["vp09", "mjpg", "h264", "h265"] = "vp09",
    scaleBar: ScaleBar = None,
    draw_frame_number=False,
    cropper=no_crop,
    filter_contours=lambda i, cont: True,
    cell_color=(255, 255, 0),
):
    """Render a video of the time-lapse.

    Args:
        imageSource (ImageSequenceSource): Your time-lapse source object.
        roiSource ([type]): Your source of RoIs for the image (e.g. cells). If None, no RoIs are visualized. Defaults to None.
        filename (str, optional): The output path of the video. Defaults to 'output.mp4'.
        framerate (int, optional): The framerate of the video. E.g. 3 means three time-lapse images per second. Defaults to 3.
        codec (str, optional): The video format codec. Defaults to "vp09".
        scaleBar (ScaleBar, optional): The scale bar object. Defaults to None.
        draw_frame_number (bool, optional): Whether to draw the frame number. Defaults to False.
        cropper ([type], optional): The frame cropper object. Defaults to no_crop.
    """

    if roiSource is None:
        # when we have no rois -> create iterator that always returns None
        def always_none():
            while True:
                yield None

        roiSource = iter(always_none())

    # make codec lower case
    codec = codec.lower()

    # create the video exporter
    ve = __video_export_from_str(filename, codec, framerate)

    with ve:
        for frame, (image, overlay) in enumerate(
            tqdm.tqdm(zip(imageSource, roiSource))
        ):

            # extract the numpy image
            if isinstance(image, BaseImage):
                image = image.raw
            elif isinstance(image, np.ndarray):
                pass
            else:
                raise Exception("Unsupported image type!")

            # copy image as we draw onto it
            image = np.copy(image)

            crop_parameters = cropper(image, overlay)
            image = image[crop_parameters[0], crop_parameters[1]]

            height, width = image.shape[:2]

            # TODO: Draw float based contours
            # Draw overlay
            if overlay:
                image = cv2.drawContours(
                    image,
                    [
                        np.array(cont.coordinates).astype(np.int32)
                        for i, cont in enumerate(
                            overlay.croppedContours(crop_parameters)
                        )
                        if filter_contours(i, cont)
                    ],
                    -1,
                    cell_color,
                )  # RGB format

            if draw_frame_number:
                cv2.putText(
                    image,
                    f"Frame: {frame}",
                    (10, 50),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    1,
                    (255, 255, 255),
                )

            if scaleBar:
                image = scaleBar.draw(
                    image, width - scaleBar.pixelWidth - 10, height - 10
                )

            # output images
            ve.write(image)


def fast_mask_rendering(masks, im, colors, alpha=0.5):
    """
    Plot masks on image.

    Args:
        masks (tensor): Predicted masks on cuda, shape: [n, h, w]
        colors (List[List[Int]]): Colors for predicted masks, [[r, g, b] * n]
        im_gpu (tensor): Image is in cuda, shape: [3, h, w], range: [0, 1]
        alpha (float): Mask transparency: 0.0 fully transparent, 1.0 opaque
        retina_masks (bool): Whether to use high resolution masks or not. Defaults to False.
    """

    colors = np.array(colors, dtype=np.float32) / 255.0  # shape(n,3)
    colors = colors[:, None, None]  # shape(n,1,1,3)

    masks = np.expand_dims(masks, 3)  # shape(n,h,w,1)
    masks_color = masks * (colors * alpha)  # shape(n,h,w,3)

    inv_alpha_masks = (1 - masks * alpha).cumprod(0)  # shape(n,h,w,1)
    mcs = masks_color.max(axis=0)  # shape(n,h,w,3)

    im = im.astype(np.float) / 255
    im = im * inv_alpha_masks[-1] + mcs
    im_mask = (im * 255).astype(np.uint8)

    return im_mask


def fast_mask_rendering_torch(masks, im, colors, alpha=0.5):
    """
    Plot masks on image.

    Args:
        masks (tensor): Predicted masks on cuda, shape: [n, h, w]
        colors (List[List[Int]]): Colors for predicted masks, [[r, g, b] * n]
        im_gpu (tensor): Image is in cuda, shape: [3, h, w], range: [0, 1]
        alpha (float): Mask transparency: 0.0 fully transparent, 1.0 opaque
        retina_masks (bool): Whether to use high resolution masks or not. Defaults to False.
    """
    # pylint: disable=import-outside-toplevel
    import torch

    colors = torch.tensor(colors, dtype=torch.float32) / 255.0  # shape(n,3)
    colors = colors[:, None, None]  # shape(n,1,1,3)

    masks = torch.tensor(masks)
    masks = masks.unsqueeze(3)  # shape(n,h,w,1)
    masks_color = masks * (colors * alpha)  # shape(n,h,w,3)

    inv_alpha_masks = (1 - masks * alpha).cumprod(0)  # shape(n,h,w,1)
    mcs = masks_color.max(dim=0).values  # shape(n,h,w,3)

    im = torch.tensor(im, dtype=torch.float) / 255
    im = im * inv_alpha_masks[-1] + mcs
    im_mask = (im * 255).byte().numpy()

    return im_mask



"""Offline image processors to perform segmentation"""

import logging

import cv2
import numpy as np
import torch
import tqdm.auto as tqdm

from acia.base import Contour, ImageSequenceSource, Overlay, Processor

from .predict import contour_from_mask, prediction


class OfflineModel(Processor):
    """
    Model that runs on the local computer
    """

    def __init__(
        self, config_file, parameter_file, half=False, device="cuda", tiling=None
    ):
        """
        config_file: model configuration file
        parameter_file: model checkpoint file
        half: enables half-precision (16-bit) execution. A bit faster.
        device: chooses the device to execute (e.g. 'cpu' or 'cuda' or 'cuda:0')
        """
        # store file destinations
        self.config_file = config_file
        self.parameter_file = parameter_file
        # empty model instance
        self.model = None
        # half-precision execution
        self.half = half
        # determine the device
        self.device = device

        self.tiling = tiling

    def load_model(self, device=None, cfg_options=None, half=False):
        """
        Load model from definitions

        device: device type, e.g. 'cpu' or 'cuda'
        cfg_options: overwrite configuration options e.g. {'test_cfg.rpn.nms_thr': 0.7}
        """
        from mmcv.runner import wrap_fp16_model
        from mmdet.apis import init_detector

        # init model
        if self.model is None:
            self.model = init_detector(
                self.config_file,
                self.parameter_file,
                device=device,
                cfg_options=cfg_options,
            )
            if half:
                # make it 16-bit
                wrap_fp16_model(self.model)

            if "classes" in self.model.cfg:
                # update object classes from config
                self.model.CLASSES = self.model.cfg["classes"]

        return self.model

    def predict(self, source: ImageSequenceSource) -> Overlay:
        """
        Predicts the overlay for an image sequence

        source: image sequence source
        tiling: whether to enable tiling
        """
        self.load_model(
            half=self.half, device=self.device
        )  # , cfg_options={'test_cfg.rcnn.nms.iou_threshold': 0.3, 'test_cfg.rcnn.score_thr': 0.5})

        # TODO: super strange without [] it takes some other list as initialization. This leads to detected cells from other images...
        overlay = Overlay([])

        for frame_id, image in tqdm.tqdm(enumerate(source)):

            pred_result = prediction(image, self.model, tiling=self.tiling)

            if len(pred_result) == 0:
                # no predictions
                continue

            all_masks = np.stack([det["mask"] for det in pred_result])
            all_contours = [contour_from_mask(mask, 0.5) for mask in all_masks]
            # drop non-sense contours
            all_contours = list(
                filter(lambda comb: len(comb[1]) >= 5, zip(pred_result, all_contours))
            )

            contours = [
                Contour(cont, pred["score"], frame_id, id=-1, label=pred["label"])
                for pred, cont in all_contours
            ]
            overlay.add_contours(contours)

        return overlay


class PoseModel(Processor):
    """
    Model that runs on the local computer
    """

    def __init__(
        self,
        model_name="bact_omni",
        omni=True,
        use_gpu=torch.cuda.is_available(),
        diameter=None,
        flow_threshold=None,
    ):
        """
        config_file: model configuration file
        parameter_file: model checkpoint file
        half: enables half-precision (16-bit) execution. A bit faster.
        device: chooses the device to execute (e.g. 'cpu' or 'cuda' or 'cuda:0')
        """
        self.omni = omni
        self.use_gpu = use_gpu
        self.model_name = model_name
        self.model = None

        self.diameter = diameter
        self.flow_threshold = flow_threshold

    def load_model(self):
        """
        Load model from definitions
        """
        from cellpose import models
        logging.info("Loading model %s", self.model_name)
        self.model = models.Cellpose(
            gpu=self.use_gpu, model_type=self.model_name, omni=self.omni
        )

    def predict(self, source: ImageSequenceSource) -> Overlay:
        """
        Predicts the overlay for an image sequence

        source: image sequence source
        tiling: whether to enable tiling
        """
        self.load_model()
        channels = [[0, 0]]

        # TODO: super strange without [] it takes some other list as initialization. This leads to detected cells from other images...
        overlay = Overlay([])

        for frame_id, image in tqdm.tqdm(enumerate(source)):
            try:
                masks, _, _, _ = self.model.eval(
                    [image],
                    channels=channels,
                    rescale=None,
                    diameter=70,
                    flow_threshold=0.9,
                    mask_threshold=0.0,
                    resample=True,
                    diam_threshold=100,
                )
            # TODO: more precise exception for Omnipose failure
            # pylint: disable=W0703
            except Exception:
                print("Error in OmniPose prediction")
                masks = [
                    [],
                ]

            int_mask = masks[0]

            num_cells = np.max(int_mask)
            score_threshold = 0.5
            all_contours = []

            for index in range(1, num_cells + 1):
                bool_mask = int_mask == index

                contours, _ = cv2.findContours(
                    np.where(bool_mask > score_threshold, 1, 0).astype(np.uint8),
                    cv2.RETR_TREE,
                    cv2.CHAIN_APPROX_SIMPLE,
                )

                for contour in contours:
                    contour = np.squeeze(contour)
                    if len(contour) > 3:
                        all_contours.append(contour)

            for contour in all_contours:
                overlay.add_contour(Contour(contour, -1, frame_id, -1))

        return overlay



"""Tracking module contains all tools to work with tracking formats
"""

from pathlib import Path

import networkx as nx
import numpy as np

from acia.base import Overlay

from .formats import gen_simple_tracking, parse_simple_tracking


class TrackingSource:
    """Base class for tracking information containing segmentation overlay and tracking graph (usually ids of overlay contours)"""

    @property
    def overlay(self) -> Overlay:
        raise NotImplementedError()

    @property
    def tracking_graph(self) -> nx.DiGraph:
        raise NotImplementedError()

    def copy(self) -> "TrackingSource":
        raise NotImplementedError()


class TrackingSourceInMemory(TrackingSource):
    """Tracking Source stored in memory"""

    def __init__(self, overlay: Overlay, tracking_graph: nx.DiGraph):
        super().__init__()
        self.__overlay = overlay
        self.__tracking_graph = tracking_graph

    @property
    def overlay(self) -> Overlay:
        return self.__overlay

    @property
    def tracking_graph(self) -> nx.DiGraph:
        return self.__tracking_graph

    def copy(self) -> "TrackingSourceInMemory":
        return TrackingSourceInMemory(
            Overlay(list(self.overlay)), self.tracking_graph.copy()
        )

    def merge(self, tr_source: TrackingSource):
        tr_source = tr_source.copy()

        self.__overlay = Overlay(self.overlay.contours + tr_source.overlay.contours)
        self.__tracking_graph = nx.compose(
            self.tracking_graph, tr_source.tracking_graph
        )

        return self


class SimpleTrackingSource(TrackingSourceInMemory):
    """Tracking Source based on simple tracking json format"""

    def __init__(self, file_content: str):
        super().__init__(*parse_simple_tracking(file_content))

    @staticmethod
    def from_file(file_path: Path) -> "SimpleTrackingSource":
        """Loads segmentation and tracking from simple tracking json format

        Args:
            file_path (Path): path to the simple tracking file

        Returns:
            SimpleTrackingSource: the loaded simple tracking file
        """
        with open(file_path, encoding="utf-8") as input_file:
            return SimpleTrackingSource(input_file.read())

    def store(self, file_path: Path):
        """Saves simple tracking json format

        Args:
            file_path (Path): file name to save
        """
        with open(file_path, "w", encoding="utf-8") as output_file:
            output_file.write(gen_simple_tracking(self.overlay, self.tracking_graph))


def subsample_tracking(
    tracking: TrackingSource, subsampling_factor: int
) -> TrackingSource:
    """Subsample the tracking source

    Args:
        tracking (TrackingSource): tracking source to subsample
        subsampling_factor (int): subsampling factor defining the step of frames. 1 means no subsampling. 2 means every second frame, ...

    Raises:
        ValueError: when wrong subsampling factor is chosen

    Returns:
        TrackingSource: subsampled tracking source
    """

    if subsampling_factor < 1:
        raise ValueError("Please chose a subsampling factor >= 1")

    # extract information from source
    overlay = tracking.overlay
    tracking_graph = tracking.tracking_graph

    # subsample frames
    subsampled_frames = set(
        np.arange(overlay.numFrames(), step=subsampling_factor, dtype=np.int32)
    )

    frame_lookup = {
        old_frame: new_frame
        for new_frame, old_frame in zip(
            range(len(subsampled_frames)), sorted(subsampled_frames)
        )
    }

    # and create overlay with remaining contours
    subsampled_overlay = Overlay(
        list(filter(lambda cont: cont.frame in subsampled_frames, overlay)),
        frames=list(range(len(subsampled_frames))),
    )

    for cont in subsampled_overlay:
        cont.frame = frame_lookup[cont.frame]

    # copy tracking graph
    subsampled_graph = tracking_graph.copy()

    # compute the set of segment ids we have to remove
    subsampled_overlay_ids = {cont.id for cont in subsampled_overlay}
    nodes_to_remove = set(tracking_graph.nodes).difference(
        subsampled_overlay_ids
    )  # [node for node in nx.topological_sort(tracking_graph) if node not in subsampled_overlay_ids]

    # loop over all these segments to remove
    for node in nodes_to_remove:
        # get parents and children
        parents = list(subsampled_graph.predecessors(node))
        children = list(subsampled_graph.successors(node))

        # for every edge: (parent --> node --> child) insert edge: (parent --> child) into the subsampled graph
        for parent in parents:
            for child in children:
                # connect parent to children
                subsampled_graph.add_edge(parent, child)

        subsampled_graph.remove_node(node)

    # make sure that we have still all contours of the overlay in our tracking
    assert len(set(subsampled_overlay_ids).difference(set(subsampled_graph.nodes))) == 0

    # return the subsampled tracking source
    return TrackingSourceInMemory(subsampled_overlay, subsampled_graph)


def ctc_track_graph(ov: Overlay, tracklet_graph: nx.DiGraph):
    """Computes the ctc track graph (every cell detection is a node) based on cell detections (overlay) and the tracklet graph (every tracklet is one node).

    Hint: overlay labels and tracklet_graph node ids need to align.

    Args:
        ov (Overlay): _description_
        tracklet_graph (nx.DiGraph): _description_

    Returns:
        _type_: _description_
    """

    track_graph = nx.DiGraph()

    # add all the nodes
    for cont in ov:
        track_graph.add_node(cont.id, frame=cont.frame)

    tracklets = {}
    for cont in ov:
        tracklets[cont.label] = tracklets.get(cont.label, []) + [cont]

    for tracklet_label in tracklets:
        tracklets[tracklet_label] = sorted(
            tracklets[tracklet_label], key=lambda c: c.frame
        )

    for tracklet_label, tracklet_nodes in tracklets.items():

        # add tracklet edges
        for contA, contB in zip(tracklet_nodes, tracklet_nodes[1:]):
            track_graph.add_edge(contA.id, contB.id)

        for pred_label in tracklet_graph.predecessors(tracklet_label):
            track_graph.add_edge(tracklets[pred_label][-1].id, tracklet_nodes[0].id)

        for succ_label in tracklet_graph.successors(tracklet_label):
            track_graph.add_edge(tracklet_nodes[-1].id, tracklets[succ_label][0].id)

    return track_graph
